@@ -31,9 +31,9 @@ import GHC.Generics (Generic)
 import System.FilePath (replaceFileName, takeBaseName, takeDirectory, (</>))
 import Text.Parsec
 import Typst.Bind (destructuringBind)
+import Typst.Constructors (getConstructor)
 import Typst.Methods (getMethod)
-import Typst.Module.Standard (loadFileText, standardModule, symModule,
-                              sysModule)
+import Typst.Module.Standard (loadFileText, standardModule, symModule)
 import Typst.Module.Math (mathModule)
 import Typst.MathClass (mathClassOf, MathClass(Relation))
 import Typst.Parse (parseTypst)
@@ -58,19 +58,25 @@ evaluateTypst ::
   -- | Markup produced by 'parseTypst'
   [Markup] ->
   m (Either ParseError Content)
-evaluateTypst operations =
+evaluateTypst operations fp =
   runParserT
     (do contents <- mconcat <$> many pContent <* eof
         -- "All documents are automatically wrapped in a document element."
         pure $ Elt "document" Nothing [("body", VContent contents)])
     initialEvalState { evalOperations = operations,
-                       evalPackageRoot = "." }
+                       evalLocalDir = takeDirectory fp }
+    fp
 
 initialEvalState :: EvalState m
 initialEvalState =
-  emptyEvalState { evalIdentifiers = [(BlockScope, standardModule')] }
+  emptyEvalState { evalIdentifiers = [(BlockScope, mempty)]
+                 , evalMathIdentifiers = [(BlockScope, mathModule <> symModule)]
+                 , evalStandardIdentifiers = [(BlockScope, standardModule'')]
+                 , evalPackageRoot = "."
+                 }
   where
     standardModule' = M.insert "eval" evalFunction standardModule
+    standardModule'' = M.insert "std" (VModule "std" standardModule') standardModule'
     evalFunction = makeFunction $ do
       code :: Text <- nthArg 1
       case parseTypst "eval" ("#{\n" <> code <> "\n}") of
@@ -240,13 +246,9 @@ pElt = do
                 [("level", VInteger (fromIntegral level))]
           }
     Equation display ms -> inBlock BlockScope $ do
-      importModule mathModule
-      importModule symModule
-      importModule sysModule
       oldMath <- evalMath <$> getState
       updateState $ \st -> st {evalMath = True}
       content <- pInnerContents ms
-      updateState $ \st -> st {evalMath = oldMath}
       element
         "equation"
         Arguments
@@ -256,7 +258,7 @@ pElt = do
                 [ ("block", VBoolean display),
                   ("numbering", VNone)
                 ]
-          }
+          } <* updateState (\st -> st {evalMath = oldMath})
     MFrac numexp denexp -> do
       let handleParens (MGroup (Just "(") (Just ")") xs) = MGroup Nothing Nothing xs
           handleParens x = x
@@ -568,6 +570,10 @@ evalExpr expr =
           arguments <- toArguments args
           applyElementFunction i (Function f) arguments
         VFunction Nothing _ (Function f) -> toArguments args >>= f
+        VType ty ->
+          case getConstructor ty of
+            Just (VFunction _ _ (Function f)) -> toArguments args >>= f
+            _ -> fail $ "No constructor defined for type " <> show ty
         VSymbol (Symbol _ True _) | mathMode ->
           do
             val' <- lookupIdentifier "accent"
@@ -817,6 +823,8 @@ evalExpr expr =
         _ -> fail $ "For expression requires an Array or Dictionary"
       updateState $ \st -> st {evalFlowDirective = FlowNormal}
       go items VNone
+    Context e -> do
+      evalExpr e -- TODO for now we just ignore "context"
     Return mbe -> do
       -- these flow directives are examined in CodeBlock
       updateState (\st -> st {evalFlowDirective = FlowReturn (isJust mbe)})
@@ -986,7 +994,6 @@ findPackageEntryPoint modname = do
 loadModule :: Monad m => Text
            -> MP m (Seq Content, (Identifier, M.Map Identifier Val))
 loadModule modname = do
-  pos <- getPosition
   (fp, modid, mbPackageRoot) <-
         if T.take 1 modname == "@"
            then do
@@ -996,13 +1003,7 @@ loadModule modname = do
                     (T.pack $ takeWhile (/= ':') . takeBaseName $
                        T.unpack modname),
                    Just (takeDirectory fp'))
-           else if T.take 1 modname == "/" -- refers to path relative to package root
-                then do
-                  packageRoot <- evalPackageRoot <$> getState
-                  pure (packageRoot </> drop 1 (T.unpack modname),
-                        Identifier (T.pack $ takeBaseName $ T.unpack modname),
-                        Nothing)
-                else pure (replaceFileName (sourceName pos) (T.unpack modname),
+           else pure (T.unpack modname,
                         Identifier (T.pack $ takeBaseName $ T.unpack modname),
                         Nothing)
   txt <- loadFileText fp
@@ -1010,6 +1011,12 @@ loadModule modname = do
     Left err -> fail $ show err
     Right ms -> do
       operations <- evalOperations <$> getState
+      currentLocalDir <- evalLocalDir <$> getState
+      let (pkgroot , localdir) =
+            case mbPackageRoot of
+              Just r -> (r , takeDirectory fp)
+              Nothing -> (evalPackageRoot initialEvalState ,
+                          currentLocalDir </> takeDirectory fp)
       res <-
         lift $
           runParserT
@@ -1020,8 +1027,8 @@ loadModule modname = do
                 pure (cs, s)
             )
             initialEvalState{evalOperations = operations,
-                             evalPackageRoot = fromMaybe (evalPackageRoot initialEvalState)
-                                                   mbPackageRoot }
+                             evalLocalDir = localdir,
+                             evalPackageRoot = pkgroot }
             fp
             ms
       case res of
@@ -1100,9 +1107,7 @@ addIdentifier ident val = do
   case identifiers of
     [] -> fail "Empty evalIdentifiers"
     ((s, i) : is) -> updateState $ \st ->
-      st
-        { evalIdentifiers = (s, M.insert ident val i) : is
-        }
+      st { evalIdentifiers = (s, M.insert ident val i) : is }
 
 updateIdentifier :: Monad m => Identifier -> Val -> MP m ()
 updateIdentifier ident val = do
