@@ -19,7 +19,7 @@ import qualified Data.Foldable as F
 import Data.List (intersperse, sort, sortOn)
 import qualified Data.Map as M
 import qualified Data.Map.Ordered as OM
-import Data.Maybe (fromMaybe)
+import Data.Maybe (fromMaybe, listToMaybe)
 import Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Vector as V
@@ -146,6 +146,8 @@ getMethod updateVal val fld = do
       case fld of
         "len" ->
           pure $ makeFunction $ pure $ VInteger (fromIntegral $ T.length t)
+        "rev" ->
+          pure $ makeFunction $ pure $ VString (T.reverse t)
         "first" ->
           if T.null t
             then fail "string is empty"
@@ -378,6 +380,12 @@ getMethod updateVal val fld = do
         "term" -> pure $ VContent t
         "description" -> pure $ VContent d
         _ -> noMethod "TermItem" fld
+    VVersion xs ->
+      case fld of
+        "at" -> pure $ makeFunction $ do
+                  i <- nthArg 1
+                  pure $ VInteger $ fromMaybe 0 $ listToMaybe $ drop i xs
+        _ -> noMethod "Version" fld
     VArray v -> do
       let toPos n =
             if n < 0
@@ -428,9 +436,12 @@ getMethod updateVal val fld = do
         "split" -> pure $ makeFunction $ do
           spliton <- nthArg 1
           let go v' = case V.break (== spliton) v' of
-                (a, b) | V.null b -> if V.null a then [] else [VArray a]
+                (a, b) | V.null b -> [VArray a | not (V.null a)]
                 (a, b) -> VArray a : go (V.drop 1 b)
           pure $ VArray $ V.fromList $ go v
+        "intersperse" -> pure $ makeFunction $ do
+          sep <- nthArg 1
+          pure $ VArray . V.fromList . intersperse sep . V.toList $ v
         "dedup" -> pure $ makeFunction $ do
           pure $ VArray $ deduplicateVector v
         "insert" -> pure $ makeFunction $ do
@@ -510,6 +521,12 @@ getMethod updateVal val fld = do
           Function fn <- nthArg 2
           let f acc y = fn Arguments {positional = [acc, y], named = OM.empty}
           lift $ foldM f start $ V.toList v
+        "reduce" -> pure $ makeFunction $ do
+          Function fn <- nthArg 1
+          let f acc y = fn Arguments {positional = [acc, y], named = OM.empty}
+          case V.toList v of
+            [] -> pure VNone
+            (x:xs) -> lift $ foldM f x xs
         "any" -> pure $ makeFunction $ do
           Function fn <- nthArg 1
           let predicate y = do
@@ -546,8 +563,30 @@ getMethod updateVal val fld = do
               VArray . V.fromList . map fst . sortOn snd
                 <$> (mapM (\x -> (x,) <$> kf' x) (V.toList v))
         "zip" -> pure $ makeFunction $ do
-          (v' :: V.Vector Val) <- ask >>= getPositionalArg 1
-          pure $ VArray $ V.map pairToArray $ V.zip v v'
+          (xs :: [Val]) <- positional <$> ask
+          let len = V.length v
+          pure $ VArray $ V.filter (/= VNone) $
+            V.map (\i -> maybe VNone (VArray . V.fromList)
+                           (mapM (\x ->
+                              case x of
+                                VArray v' -> v' V.!? i
+                                _ -> Nothing) (val : xs)))
+              (V.enumFromTo 0 (len - 1))
+        "to-dict" -> pure $ makeFunction $
+          VDict . OM.fromList <$>
+                 mapM (\x -> do
+                         vx <- fromVal x
+                         case V.toList vx of
+                           [a,b] -> do
+                             k <- fromVal a
+                             pure (Identifier k, b)
+                           _ -> fail "vector has wrong shape") (V.toList v)
+        "windows" -> pure $ makeFunction $ do
+          (windowsize :: Int) <- nthArg 1
+          case V.length v - windowsize of
+            n | n < 0 -> pure $ VArray mempty
+              | otherwise -> pure $ VArray $ V.fromList $
+                   map (\x -> VArray $ V.take windowsize $ V.drop x v) [0..n]
         "sum" -> pure $ makeFunction $ do
           mbv <- namedArg "default" Nothing
           case V.uncons v of
@@ -620,6 +659,21 @@ getMethod updateVal val fld = do
     VArguments args ->
       case fld of
         "pos" -> pure $ makeFunction $ pure $ VArray $ V.fromList (positional args)
+        "at" ->
+          pure $ makeFunction $ do
+            (x :: Val) <- nthArg 1
+            defval <- namedArg "default" VNone
+            case x of
+              VInteger{} -> do
+                i <- fromVal x
+                case positional args of
+                  xs | i < length xs -> pure $ xs !! i
+                     | otherwise -> pure defval
+              VString t ->
+                case OM.lookup (Identifier t) (named args) of
+                  Just a -> pure a
+                  Nothing -> pure defval
+              _ -> pure defval
         "named" -> pure $ makeFunction $ pure $ VDict $ named args
         _ -> noMethod "Arguments" fld
     VDateTime mbdate mbtime -> do
@@ -671,9 +725,6 @@ getMethod updateVal val fld = do
         _ -> noMethod "DateTime" fld
     _ -> noMethod (drop 1 $ takeWhile (/= ' ') $ show val) fld
 
-pairToArray :: (Val, Val) -> Val
-pairToArray (x, y) = VArray $ V.fromList [x, y]
-
 formatNumber :: Text -> Int -> Text
 formatNumber t n = F.foldMap go $ T.unpack t
   where
