@@ -40,7 +40,8 @@ data PState = PState
   { stIndent :: [Int],
     stLineStartCol :: !Int,
     stAllowNewlines :: !Int, -- allow newlines if > 0
-    stBeforeSpace :: Maybe (SourcePos, Text),
+    stSpaceBefore :: Maybe (SourcePos, Text),
+    stLastMathTok :: Maybe (SourcePos, Markup),
     stContentBlockNesting :: Int
   }
   deriving (Show)
@@ -51,7 +52,8 @@ initialState =
     { stIndent = [],
       stLineStartCol = 1,
       stAllowNewlines = 0,
-      stBeforeSpace = Nothing,
+      stSpaceBefore = Nothing,
+      stLastMathTok = Nothing,
       stContentBlockNesting = 0
     }
 
@@ -69,9 +71,9 @@ ws = do
         | allowNewlines > 0 = isSpace c
         | otherwise = isSpace c && c /= '\r' && c /= '\n'
   ( skipMany1 (void (satisfy isSp) <|> void pComment)
-      *> updateState (\st -> st {stBeforeSpace = Just (p1, inp)})
+      *> updateState (\st -> st {stSpaceBefore = Just (p1, inp)})
     )
-    <|> updateState (\st -> st {stBeforeSpace = Nothing})
+    <|> updateState (\st -> st {stSpaceBefore = Nothing})
 
 lexeme :: P a -> P a
 lexeme pa = pa <* ws
@@ -164,32 +166,52 @@ pEquation = do
 
 mathOperatorTable :: [[Operator Text PState Identity Markup]]
 mathOperatorTable =
-  [ -- precedence 7 -- attachment with number, e.g. a_1 (see #17)
-    [ Infix (attachBottom <$ (try (op "_" *> lookAhead mNumber))) AssocLeft,
-      Infix (attachTop <$ (try (op "^" *> lookAhead mNumber))) AssocLeft
+  [ -- precedence 7 -- attachment with number, e.g. a_1 (#17), or (..) group
+    [ Infix (attachBottom <$ (try (op "_" *> lookAhead (mNumber <|> mGroup))))
+        AssocLeft,
+      Infix (attachTop <$ (try (op "^" *> lookAhead (mNumber <|> mGroup))))
+        AssocLeft
     ],
     -- precedence 6
     [ Postfix
         ( try $ do
-            mbBeforeSpace <- stBeforeSpace <$> getState
+            getState >>= guard . isNothing . stSpaceBefore
             -- NOTE: can't have space before () or [] arg in a
             -- function call! to prevent bugs with e.g. 'if 2<3 [...]'.
-            guard $ isNothing mbBeforeSpace
+            pos <- getPosition
+            lastMathTok <- stLastMathTok <$> getState
+            -- 1(a) is not a function
+            -- !(a) is not a function
+            -- f(a) is a function
+            -- "alpha"(a) is a function
+            -- alpha(a) is a function
+            -- see #55
+            -- but we still don't match typst for "!"(a), which typst DOES consider
+            -- a function
+            guard $ case lastMathTok of
+                      Just (pos', MGroup _ (Just t) _)
+                        | pos == pos' -> T.all isLetter t
+                      Just (pos', Text t)
+                        | pos == pos'
+                        -> case T.unsnoc t of
+                                  Nothing -> True
+                                  Just (_,c) -> isLetter c
+                      _ -> True
             args <- mGrouped '(' ')' True
             pure $ \expr -> MGroup Nothing Nothing [expr, args]
         )
     ],
-    -- precedence 5 -- attachment with non-number, e.g. a_x
-    [ Infix (attachBottom <$ op "_") AssocLeft,
-      Infix (attachTop <$ op "^") AssocLeft
-    ],
-    -- precedence 4  -- factorial needs to take precedence over fraction
+    -- precedence 5  -- factorial needs to take precedence over fraction
     [ Postfix (try $ do
-                  mbBeforeSpace <- stBeforeSpace <$> getState
+                  mbBeforeSpace <- stSpaceBefore <$> getState
                   guard $ isNothing mbBeforeSpace
                   lexeme $ char '!' *> notFollowedBy (char '=')
                   pure (\expr -> MGroup Nothing Nothing [expr, Text "!"]))
     ],
+    -- precedence 4 -- attachment with non-number, e.g. a_x
+    [ Infix (attachBottom <$ op "_") AssocLeft,
+      Infix (attachTop <$ op "^") AssocLeft
+    ],
     -- precedence 3
     [ Infix (makeFrac <$ op "/") AssocLeft
     ]
@@ -231,7 +253,7 @@ mathFunctionCall :: Operator Text PState Identity Expr
 mathFunctionCall =
   Postfix
     ( do
-        mbBeforeSpace <- stBeforeSpace <$> getState
+        mbBeforeSpace <- stSpaceBefore <$> getState
         -- NOTE: can't have space before () or [] arg in a
         -- function call! to prevent bugs with e.g. 'if 2<3 [...]'.
         guard $ isNothing mbBeforeSpace
@@ -275,7 +297,8 @@ pMath :: P Markup
 pMath = buildExpressionParser mathOperatorTable pBaseMath
 
 pBaseMath :: P Markup
-pBaseMath = mNumber
+pBaseMath = do
+  tok <-    mNumber
         <|> mLiteral
         <|> mEscaped
         <|> mShorthand
@@ -286,6 +309,9 @@ pBaseMath = mNumber
         <|> mCode
         <|> mMid
         <|> mSymbol
+  pos <- getPosition
+  updateState $ \s -> s{ stLastMathTok = Just (pos, tok) }
+  pure tok
 
 mGroup :: P Markup
 mGroup = mGrouped '(' ')' False
@@ -307,17 +333,17 @@ mNumber = lexeme $ do
       mempty
       ( do
           e <- char '.'
-          es <- many1 digit
+          es <- many digit
           pure $ T.pack (e : es)
       )
   pure $ Text (ds <> opt)
 
 mLiteral :: P Markup
 mLiteral = do
-  mbBeforeSpace <- stBeforeSpace <$> getState
+  mbBeforeSpace <- stSpaceBefore <$> getState
   String t <- pStr
   -- ensure space in e.g. x "is natural":
-  mbAfterSpace <- stBeforeSpace <$> getState
+  mbAfterSpace <- stSpaceBefore <$> getState
   pure $
     Text $
       (maybe "" (const " ") mbBeforeSpace)
@@ -384,7 +410,7 @@ mCode = lexeme $ char '#' *> (Code <$> getPosition <*> pBasicExpr)
 
 mMid :: P Markup
 mMid = try $ do
-  getState >>= guard . isJust . stBeforeSpace
+  getState >>= guard . isJust . stSpaceBefore
   void $ char '|' *> space *> ws
   pure $ MGroup Nothing Nothing [Nbsp, Text "|", Nbsp]
 
@@ -588,11 +614,50 @@ pText :: P Markup
 pText = Text . mconcat <$> some
   ((do xs <- some alphaNum
        T.pack . (xs <>) <$>
-             try (some (char '*' <|> char '_') <* lookAhead alphaNum)
+             try (some (char '*' <|> char '_') <* lookAhead (satisfy nonCJKAlphaNum))
         <|> pure (T.pack xs))
  <|> (T.pack <$> some (satisfy (\c -> not (isSpace c || isSpecial c))))
   )
 
+nonCJKAlphaNum :: Char -> Bool
+nonCJKAlphaNum c = isAlphaNum c && not (isCJK c)
+
+isCJK :: Char -> Bool
+isCJK c | c < '\x2e80' = False
+isCJK c =
+          (c >= '\x2e80' && c <= '\x2eff') -- CJK Radicals Supplement
+       || (c >= '\x2f00' && c <= '\x2fdf') -- Kangxi Radicals
+       || (c >= '\x2ff0' && c <= '\x2fff') -- Ideographic Description Characters
+       || (c >= '\x3000' && c <= '\x303f') -- JK Symbols and Punctuation
+       || (c >= '\x3040' && c <= '\x309f') -- Hiragana
+       || (c >= '\x30a0' && c <= '\x30ff') -- Katakana
+       || (c >= '\x3100' && c <= '\x312f') -- Bopomofo
+       || (c >= '\x3130' && c <= '\x318f') -- Kanbun
+       || (c >= '\x3190' && c <= '\x319f') -- Kanbun
+       || (c >= '\x31c0' && c <= '\x31ef') -- CJK Strokes
+       || (c >= '\x31f0' && c <= '\x31ff') -- Katakana Phonetic Extensions
+       || (c >= '\x3200' && c <= '\x32ff') -- Enclosed CJK Letters & Months
+       || (c >= '\x3300' && c <= '\x33ff') -- CJK Compatibility
+       || (c >= '\x3400' && c <= '\x4dbf') -- CJK Unified Ideographs Extension A
+       || (c >= '\x4e00' && c <= '\x9fff') -- CJK Unified Ideographs
+       || (c >= '\xa000' && c <= '\xa48f') -- Yi Syllables
+       || (c >= '\xa490' && c <= '\xa4cf') -- Yi Radicals
+       || (c >= '\xf900' && c <= '\xfaff') -- CJK Compatibility Ideographs
+       || (c >= '\xfe10' && c <= '\xfe1f') -- Vertical forms
+       || (c >= '\xfe30' && c <= '\xfe4f') -- CJK Compatibility Forms
+       || (c >= '\xFE50' && c <= '\xFE6F') -- Small Form Variants
+       || (c >= '\xFF00' && c <= '\xFFEE') -- Halfwidth and Fullwidth Forms
+       || (c >= '\x1B000' && c <= '\x1B0FF') -- Kana Supplement
+       || (c >= '\x1B100' && c <= '\x1B12F') -- Kana Extended-A
+       || (c >= '\x1B130' && c <= '\x1B16F') -- Small Kana Extension
+       || (c >= '\x20000' && c <= '\x2A6DF') -- CJK Unified Ideographs Extension B
+       || (c >= '\x2A700' && c <= '\x2B73F') -- CJK Unified Ideographs Extension C
+       || (c >= '\x2B740' && c <= '\x2B81F') -- CJK Unified Ideographs Extension D
+       || (c >= '\x2B820' && c <= '\x2CEAF') -- CJK Unified Ideographs Extension E
+       || (c >= '\x2CEB0' && c <= '\x2EBEF') -- CJK Unified Ideographs Extension F
+       || (c >= '\x2F800' && c <= '\x2FA1F') -- CJK Compatibility Ideographs Supp
+       || (c >= '\x30000' && c <= '\x3134F') -- CJK Unified Ideographs Exten
+
 pEscaped :: P Markup
 pEscaped = Text . T.singleton <$> pEsc
 
@@ -648,12 +713,12 @@ pLabelInContent = Code <$> getPosition <*> pLabel
 pLabel :: P Expr
 pLabel =
   Label . T.pack
-    <$> try
+    <$> lexeme (try
       ( char '<'
           *> many1 (satisfy isIdentContinue <|>
                     char '_' <|> char '.' <|> char ':')
           <* char '>'
-      )
+      ))
 
 pRef :: P Markup
 pRef =
@@ -671,7 +736,7 @@ pHash = do
   void $ char '#'
   res <- Code <$> getPosition <*> pBasicExpr <* optional (sym ";")
   -- rewind if we gobbled space:
-  mbBeforeSpace <- stBeforeSpace <$> getState
+  mbBeforeSpace <- stSpaceBefore <$> getState
   case mbBeforeSpace of
     Nothing -> pure ()
     Just (pos, inp) -> do
@@ -792,7 +857,7 @@ functionCall :: Operator Text PState Identity Expr
 functionCall =
   Postfix
     ( do
-        mbBeforeSpace <- stBeforeSpace <$> getState
+        mbBeforeSpace <- stSpaceBefore <$> getState
         -- NOTE: can't have space before () or [] arg in a
         -- function call! to prevent bugs with e.g. 'if 2<3 [...]'.
         guard $ isNothing mbBeforeSpace
@@ -951,11 +1016,13 @@ pContent = do
   void $ char '['
   col <- sourceColumn <$> getPosition
   oldLineStartCol <- stLineStartCol <$> getState
+  oldIndent <- stIndent <$> getState
   updateState $ \st ->
     st
       { stLineStartCol = col,
         stContentBlockNesting =
-          stContentBlockNesting st + 1
+          stContentBlockNesting st + 1,
+        stIndent = []
       }
   ms <- manyTill pMarkup (char ']')
   ws
@@ -963,7 +1030,8 @@ pContent = do
     st
       { stLineStartCol = oldLineStartCol,
         stContentBlockNesting =
-          stContentBlockNesting st - 1
+          stContentBlockNesting st - 1,
+        stIndent = oldIndent
       }
   pure $ Content ms
 
@@ -1025,6 +1093,7 @@ pKeywordExpr =
     <|> pBreakExpr
     <|> pContinueExpr
     <|> pReturnExpr
+    <|> pContextExpr
 
 -- args ::= ('(' (arg (',' arg)* ','?)? ')' content-block*) | content-block+
 pArgs :: P [Arg]
@@ -1033,7 +1102,7 @@ pArgs = do
   args <- option [] $ inParens $ sepEndBy pArg (sym ",")
   blocks <- many $ do
     -- make sure we haven't had a space
-    skippedSpaces <- isJust . stBeforeSpace <$> getState
+    skippedSpaces <- isJust . stSpaceBefore <$> getState
     if skippedSpaces
       then mzero
       else do
@@ -1167,10 +1236,11 @@ pImportExpr = pKeyword "import" *> (Import <$> pExpr <*> pImportItems)
     pImportItems =
         (sym ":"
           *> ( (AllIdentifiers <$ sym "*")
-                 <|> (SomeIdentifiers
-                       <$> sepEndBy pIdentifierAs (sym ","))
+                 <|> inParens pIdentifierList
+                 <|> pIdentifierList
              )
         ) <|> (NoIdentifiers <$> pAs)
+    pIdentifierList = SomeIdentifiers <$> sepEndBy pIdentifierAs (sym ",")
     pIdentifierAs = do
       ident <- pIdentifier
       mbAs <- pAs
@@ -1192,5 +1262,8 @@ pReturnExpr = do
     then pure $ Return Nothing
     else Return <$> (option Nothing (Just <$> pExpr))
 
+pContextExpr :: P Expr
+pContextExpr = Context <$> (pKeyword "context" *> pExpr)
+
 pIncludeExpr :: P Expr
 pIncludeExpr = Include <$> (pKeyword "include" *> pExpr)
